"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getOverriddenFunction = exports.parseArgs = exports.functionsParser = exports.getFunctionObjectSignature = exports.getFunctionInlineSignature = void 0;
const types_1 = require("./types");
const parseSchemas_1 = __importDefault(require("./parseSchemas"));
const definitions_1 = __importDefault(require("./definitions"));
const embeddedDefinitions = (0, parseSchemas_1.default)(definitions_1.default);
function compareArgumentPosition(a, b) {
    return (a.position ?? Infinity) - (b.position ?? Infinity);
}
const getFunctionInlineSignature = (name, func, requiredOnly) => {
    return ("$$" +
        name +
        (func.arguments?.some(a => a.position === 0 && a.required)
            ? "(" +
                func.arguments
                    .filter(a => typeof a.position === "number" && (!requiredOnly || a.required))
                    .sort(compareArgumentPosition)
                    .map(a => (a.required ? "{" + a.name + "}" : "[" + a.name + "]"))
                    .join(",") +
                ")"
            : ""));
};
exports.getFunctionInlineSignature = getFunctionInlineSignature;
const getFunctionObjectSignature = (name, func) => {
    return `{ "\\$\\$${name}": {input} ${func.arguments?.some(a => a.position === 0 && a.required)
        ? "," +
            func.arguments
                .filter(a => typeof a.position === "number" && a.required)
                .sort(compareArgumentPosition)
                .map(a => `"${a.name}": {${a.name}}`)
                .join(", ")
        : ""}}`;
};
exports.getFunctionObjectSignature = getFunctionObjectSignature;
class FunctionsParser {
    constructor() {
        /**
         * Match Groups: DO NOT CHANGE
         * 0 - Everything
         * 1 - Function name (no $$)
         * 2 - Args with parenthesis
         * 3 - Args without parenthesis
         * 4 - Colon (or something else)
         * 5 - What comes after the colon (value) -- only when 'noArgs' = false
         */
        this.inlineFunctionRegexFactory = (global, noArgs) => new RegExp(`\\$\\$(${this.getNames().join("|")})(\\((.*?)\\))?(:${noArgs ? "" : `([^"]*)`}|$|")`, global ? "g" : undefined);
        this.objectFunctionRegexFactory = () => new RegExp(`(?<=")\\$\\$(${this.getNames().join("|")})":`, "g");
        this.clientFunctions = {};
        this.allFunctionsNames = types_1.EmbeddedTransformerFunctions;
        this.objectFunctionRegex = this.objectFunctionRegexFactory();
        this.inlineFunctionRegex = this.inlineFunctionRegexFactory(true, true);
        this.matchInlineFunctionWithArgsRegex = this.inlineFunctionRegexFactory(false, false);
    }
    setClientFunctions(clientFunctions, handler, docsUrlResolver) {
        this.clientFunctions = (0, parseSchemas_1.default)(clientFunctions, true);
        this.clientDocsUrlResolver = docsUrlResolver;
        this.allFunctionsNames = Object.keys(clientFunctions).concat(types_1.EmbeddedTransformerFunctions).sort();
        this.objectFunctionRegex = this.objectFunctionRegexFactory();
        this.inlineFunctionRegex = this.inlineFunctionRegexFactory(true, true);
        this.matchInlineFunctionWithArgsRegex = this.inlineFunctionRegexFactory(false, false);
        this.handleClientFunction = handler;
    }
    get(name, args) {
        return this.clientFunctions?.[name] ?? embeddedDefinitions[name];
    }
    getNames() {
        return this.allFunctionsNames;
    }
    resolveDocsUrl(funcName, functionDescriptor) {
        const name = functionDescriptor?.aliasTo ?? funcName;
        if (functionDescriptor?.custom && this.clientDocsUrlResolver) {
            const url = this.clientDocsUrlResolver(name);
            if (url) {
                return url;
            }
        }
        return `https://nlighten-oss.github.io/json-transform/functions/${name}`;
    }
    matchInline(data, callback) {
        if (typeof data !== "string")
            return null;
        const m = data.match(this.matchInlineFunctionWithArgsRegex);
        if (!m) {
            return null;
        }
        const funcName = m[1];
        let func = exports.functionsParser.get(funcName);
        if (!func)
            return null;
        if (func.overrides || callback) {
            const argsWithoutParenthesis = m[3];
            const args = (0, exports.parseArgs)(func, argsWithoutParenthesis);
            func = (0, exports.getOverriddenFunction)(func, args);
            if (callback) {
                const inlineFunctionValue = m[5];
                callback(funcName, func, inlineFunctionValue, args);
            }
        }
        return func;
    }
    matchObject(data, extractOutputTypeRecursively) {
        if (!data || typeof data !== "object")
            return;
        for (const funcName of this.allFunctionsNames) {
            const key = "$$" + funcName;
            const value = data[key];
            if (typeof value !== "undefined") {
                if (extractOutputTypeRecursively && this.get(funcName).pipedType) {
                    const match = this.matchObject(value, true);
                    if (match)
                        return match;
                }
                const func = (0, exports.getOverriddenFunction)(this.get(funcName), data);
                return {
                    name: funcName,
                    func,
                    value,
                    spec: data,
                };
            }
        }
    }
    matchAllObjectFunctionsInLine(line) {
        return line.matchAll(this.objectFunctionRegex);
    }
    matchAllInlineFunctionsInLine(line) {
        return line.matchAll(this.inlineFunctionRegex);
    }
}
exports.functionsParser = new FunctionsParser();
const parseArgs = (func, args) => {
    if (!args)
        return {};
    const argsValues = args.split(/,(?=(?:[^']*'[^']*')*[^']*$)/).map((argVal) => {
        let arg = argVal;
        const trimmed = argVal.trim();
        if (trimmed.startsWith("'") && trimmed.endsWith("'") && trimmed.length > 1) {
            arg = trimmed.substring(1, trimmed.length - 1).replace(/''/g, "'");
            //otherwise, take the whole argument as-is
        }
        return arg;
    });
    return !argsValues || argsValues.length === 0
        ? {}
        : argsValues.reduce((a, c, i) => {
            func.arguments?.forEach(fa => {
                if (fa.position === i && fa.name) {
                    a[fa.name] = c;
                }
            });
            return a;
        }, {});
};
exports.parseArgs = parseArgs;
const getOverriddenFunction = (func, args) => {
    if (!func.overrides || !args)
        return func;
    for (const override of func.overrides) {
        if (override.if.every(c => (args[c.argument] ?? override.then.defaultValues?.[c.argument])?.toString().toUpperCase() === c.equals)) {
            return override.then; // replace func instance based on args
        }
    }
    return func;
};
exports.getOverriddenFunction = getOverriddenFunction;
//# sourceMappingURL=functionsParser.js.map